package j80.spectrum;

import j80.*;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import java.util.HashMap;

import java.io.*;

/**
 * $Id: Spectrum48K.java,v 1.2 2005/03/18 16:40:48 mviara Exp $
 *
 * Sinclair spectrum emulator
 *
 * $Log: Spectrum48K.java,v $
 * Revision 1.2  2005/03/18 16:40:48  mviara
 * Added support for speaker.
 *
 * Revision 1.1  2004/06/20 16:25:58  mviara
 * Split spectrum emulator in more files.
 * Added partial support for 128K.
 *
 */
public class Spectrum48K implements Peripheral,MMU,Polling,Spectrum
{




	// Connected CPU
	protected J80 cpu;

	protected Screen screen;
	protected ZXSnapshot snapshot;
	private byte memory[];

	public Spectrum48K()
	{
		screen = new Screen();
		snapshot = new ZXSnapshot();
		memory = new byte[0x10000];
	}

	// J80.MMU
	public void pokeb(int add,int value)
	{
		if (cpu.isRunning() && add < RAM_MEMORY_START)
			return;

		memory[add] = (byte)value;

		if (add >= SCREEN_MEMORY_START && add <= SCREEN_MEMORY_END)
			screen.repaintScreen(add - SCREEN_MEMORY_START);
		else if (add >= SCREEN_ATTRIBUTE_START && add <= SCREEN_ATTRIBUTE_END)
			screen.repaintAttribute(add - SCREEN_ATTRIBUTE_START);
	}

	public int peekb(int add)
	{
		return memory[add] & 0xff;
	}

	public void resetCPU(J80 cpu) throws Exception
	{
		cpu.addPeripheral(screen);
		cpu.addPeripheral(snapshot);
	}

	public void disconnectCPU(J80 cpu)
	{
	}


	public void connectCPU(J80 cpu) throws Exception
	{
		this.cpu = cpu;
		screen.setScreenMemory(memory,SCREEN_MEMORY_START);
		cpu.addPolling(20,this);

	}

	/**
	 * Polling called every 20 ms
	 */
	public void polling(J80 cpu)
	{
		cpu.irq();


	}





	public String toString()
	{
		return "Spectrum 48K $Revision: 1.2 $";
	}

}
